# -*- coding: utf-8 -*-
import unittest

import mock


class TestFacebookCanvasDecorator(unittest.TestCase):

    def test_init(self):
        from pyramid_facebook import facebook_canvas
        from pyramid_facebook.security import ViewCanvas

        dec = facebook_canvas(**{})

        self.assertEqual(ViewCanvas, dec.permission)
        self.assertEqual('facebook_canvas', dec.route_name)


class TestFacebookPaymentsGetItemsDecorator(unittest.TestCase):

    @mock.patch('pyramid_facebook.venusian')
    def test_call(self, m_venusian):
        from pyramid_facebook import facebook_payments_get_items
        wrapped = mock.Mock()

        dec = facebook_payments_get_items()
        res = dec(wrapped)

        m_venusian.attach.assert_called_once_with(wrapped, dec._register)
        self.assertEqual(wrapped, res)

    def test_register(self):
        from pyramid_facebook.security import ViewCanvas
        from pyramid_facebook import facebook_payments_get_items
        wrapped = mock.Mock()
        scanner = mock.Mock()
        scanner.config = config = mock.Mock()

        dec = facebook_payments_get_items()
        dec._decorate = mock.Mock()

        dec._register(scanner, 'name', wrapped)


        config.add_view.assert_called_once_with(
            view=dec._decorate.return_value,
            permission=ViewCanvas,
            route_name='facebook_payments_get_items',
            renderer='json',
            )

    def test_validate(self):
        from pyramid_facebook import facebook_payments_get_items

        dec = facebook_payments_get_items()

        self.assertRaises(TypeError, dec._validate, ['not_a_dict'])

        self.assertRaises(TypeError, dec._validate, {'not': 'good_keys'})

        bad_type = {
            'title'      : u"A title",
            'description': u"A description",
            'price'      : 1.2,  # not an int
            'image_url'  : "can be a wrong url :-(",
            }
        self.assertRaises(TypeError, dec._validate, bad_type)

        good_content =  {
            'title'      : u"A title",
            'description': u"A description",
            'price'      : 100,
            'image_url'  : u"any str or unicode",
            }

        self.assertEqual(good_content, dec._validate(good_content))

    def test_decorate(self):
        from pyramid_facebook import facebook_payments_get_items
        context = mock.MagicMock()
        request = mock.Mock()
        wrapped = mock.Mock()

        dec = facebook_payments_get_items()
        dec._validate = mock.Mock()

        wrapper = dec._decorate(wrapped)

        # test wrapper call:
        res = wrapper(context, request)

        # assertions:
        dec._validate.assert_called_once_with(wrapped.return_value)
        wrapped.assert_called_once_with(
            context,
            request
            )
        self.assertEqual(res, {
            "content": [dec._validate.return_value],
            "method": "payments_get_items"
            })


class TestPyramidFacebook(unittest.TestCase):

    @mock.patch('pyramid_facebook.request_params_predicate')
    @mock.patch('pyramid_facebook.FacebookAuthenticationPolicy')
    @mock.patch('pyramid_facebook.ACLAuthorizationPolicy')
    def test_includeme(self, m_acl_policy, m_fb_policy, m_predicate):
        settings = {
            'facebook.namespace': 'namespace',
            'facebook.secret_key': 'secret_key'
            }

        config = mock.MagicMock()
        config.registry.settings = settings

        # TEST includeme
        from pyramid_facebook import includeme
        includeme(config)

        # now we check if everything went as expected:
        m_fb_policy.assert_called_once_with('secret_key')
        m_acl_policy.assert_called_once_with()

        config.set_authentication_policy.assert_called_once_with(
            m_fb_policy.return_value
            )
        config.set_authorization_policy.assert_called_once_with(
            m_acl_policy.return_value
            )
        config.scan.assert_called_once_with('pyramid_facebook.views')

        self.assertEqual(5, m_predicate.call_count)

        self.assertEqual(
            mock.call(
                'signed_request',
                'buyer',
                'receiver',
                'order_id',
                'order_info',
                method='payments_get_items'
                ),
            m_predicate.call_args_list[0]
            )

        self.assertEqual(
            mock.call(
                'signed_request',
                'order_details',
                'order_id',
                method='payments_status_update',
                ),
            m_predicate.call_args_list[1]
            )

        self.assertEqual(
            mock.call(status='placed'),
            m_predicate.call_args_list[2]
            )

        self.assertEqual(
            mock.call(status='disputed'),
            m_predicate.call_args_list[3]
            )

        self.assertEqual(
            mock.call(status='refunded'),
            m_predicate.call_args_list[4]
            )

        self.assertEqual(6, config.add_route.call_count)

        call = config.add_route.call_args_list[0]
        expected = mock.call(
            'facebook_canvas',
            '/%s/' % settings['facebook.namespace'],
            request_method='POST',
            request_param='signed_request',
            factory='pyramid_facebook.security.SignedRequestContext'
            )
        self.assertEqual(expected, call)

        call = config.add_route.call_args_list[1]
        expected = mock.call(
            'facebook_canvas_oauth',
            '/%s/oauth' % settings['facebook.namespace'],
            request_method='POST',
            request_param='signed_request',
            factory='pyramid_facebook.security.SignedRequestContext'
            )
        self.assertEqual(expected, call)

        call = config.add_route.call_args_list[2]
        expected = mock.call(
            'facebook_payments_get_items',
            '/%s/credits' % settings['facebook.namespace'],
            request_method='POST',
            custom_predicates=[m_predicate.return_value],
            factory='pyramid_facebook.security.FacebookCreditsContext',
            )
        self.assertEqual(expected, call)

        call = config.add_route.call_args_list[3]
        expected = mock.call(
            'facebook_payments_status_update_placed',
            '/%s/credits' % settings['facebook.namespace'],
            request_method='POST',
            custom_predicates=[m_predicate.return_value,m_predicate.return_value],
            factory='pyramid_facebook.security.FacebookCreditsContext',
            )
        self.assertEqual(expected, call)

        call = config.add_route.call_args_list[4]
        expected = mock.call(
            'facebook_payments_status_update_disputed',
            '/%s/credits' % settings['facebook.namespace'],
            request_method='POST',
            custom_predicates=[m_predicate.return_value,m_predicate.return_value],
            factory='pyramid_facebook.security.FacebookCreditsContext',
            )
        self.assertEqual(expected, call)

        call = config.add_route.call_args_list[5]
        expected = mock.call(
            'facebook_payments_status_update_refunded',
            '/%s/credits' % settings['facebook.namespace'],
            request_method='POST',
            custom_predicates=[m_predicate.return_value,m_predicate.return_value],
            factory='pyramid_facebook.security.FacebookCreditsContext',
            )
        self.assertEqual(expected, call)


    @mock.patch('pyramid_facebook.FacebookAuthenticationPolicy')
    @mock.patch('pyramid_facebook.ACLAuthorizationPolicy')
    def test_includeme_exception(self, m_acl_policy, m_fb_policy):
        from pyramid_facebook import includeme
        bad_settings = {}

        config = mock.MagicMock()
        config.registry.settings = bad_settings

        self.assertRaises(KeyError, includeme, config)

        bad_settings = {
            'facebook.secret_key': 'secret_key'
        }

        config.registry.settings = bad_settings
        self.assertRaises(KeyError, includeme, config)



# -*- coding: utf-8 -*-
import unittest

import mock
from pyramid.security import(
    Authenticated,
    Everyone,
    Allow
    )


class TestFacebookCreditsContext(unittest.TestCase):

    @mock.patch('pyramid_facebook.security.json')
    def test_order_details(self, m_json):
        from pyramid_facebook.security import FacebookCreditsContext
        ctx = FacebookCreditsContext(None)
        ctx._fb_data = data = mock.MagicMock()

        self.assertEqual(m_json.loads.return_value, ctx.order_details)
        self.assertEqual(m_json.loads.return_value, ctx._order_details)

        m_json.loads.assert_called_once_with(data['credits']['order_details'])

    @mock.patch('pyramid_facebook.security.json')
    def test_earned_currency_data(self, m_json):
        from pyramid_facebook.security import FacebookCreditsContext
        ctx = FacebookCreditsContext(None)
        ctx._fb_data = data = mock.MagicMock()

        self.assertEqual(
            m_json.loads.return_value['modified'],
            ctx.earned_currency_data
            )
        self.assertEqual(
            m_json.loads.return_value['modified'],
            ctx._earned_currency_data
            )

    def test_order_info(self):
        from pyramid_facebook.security import FacebookCreditsContext
        ctx = FacebookCreditsContext(None)
        ctx._fb_data = data = mock.MagicMock()

        self.assertEqual(ctx._fb_data["credits"]["order_info"], ctx.order_info)



class TestSignedRequestContext(unittest.TestCase):

    def test_init(self):
        from pyramid_facebook.security import(
            SignedRequestContext,
            ViewCanvas
            )

        ctx = SignedRequestContext(None)
        self.assertListEqual(
            [(Allow, Authenticated, ViewCanvas)],
            ctx.__acl__
            )

    def test_facebook_data(self):
        from pyramid_facebook.security import(
            SignedRequestContext,
            ViewCanvas
            )

        ctx = SignedRequestContext(None)
        self.assertIsNone(ctx._fb_data)
        self.assertIsNone(ctx.facebook_data)
        ctx.facebook_data = "some data"
        self.assertEqual(
            "some data",
            ctx.facebook_data
            )

        self.assertRaises(
            ValueError,
            ctx.__setattr__,
            'facebook_data',
            "some more data"
            )

    def test_repr(self):
        from pyramid_facebook.security import (SignedRequestContext,
            FacebookCreditsContext)
        self.assertEqual(
            "<SignedRequestContext facebook_data=None>",
            repr(SignedRequestContext(None))
            )
        self.assertEqual(
            "<FacebookCreditsContext facebook_data=None>",
            repr(FacebookCreditsContext(None))
            )
class TestFacebookAuthenticationPolicy(unittest.TestCase):

    def test_init(self):
        from pyramid_facebook.security import FacebookAuthenticationPolicy

        fap = FacebookAuthenticationPolicy('secret key', 'callback')
        self.assertEqual('secret key', fap.secret_key)
        self.assertEqual('callback', fap.callback)

    def test_unauthenticated_userid_no_signed_request(self):
        from pyramid_facebook.lib import encrypt_signed_request
        from pyramid_facebook.security import FacebookAuthenticationPolicy

        # no signed request
        fap = FacebookAuthenticationPolicy('secret key')
        request = mock.MagicMock()
        request.params = dict()

        self.assertIsNone(fap.unauthenticated_userid(request))

    def test_unauthenticated_userid_no_user_id_in_signed_request(self):
        from pyramid_facebook.lib import encrypt_signed_request
        from pyramid_facebook.security import (
            FacebookAuthenticationPolicy,
            SignedRequestContext,
            )

        # no user_id in signed request
        fap = FacebookAuthenticationPolicy('secret key')
        signed_request = encrypt_signed_request(
            'secret key',
            {}
            )
        request = mock.MagicMock()
        request.params = dict(signed_request=signed_request)
        request.context = SignedRequestContext(request)

        self.assertIsNone(fap.unauthenticated_userid(request))

    def test_unauthenticated_userid_in_signed_request(self):
        from pyramid_facebook.lib import encrypt_signed_request
        from pyramid_facebook.security import (
            FacebookAuthenticationPolicy,
            SignedRequestContext,
            )

        # user id in signed request
        fap = FacebookAuthenticationPolicy('secret key')
        signed_request = encrypt_signed_request(
            'secret key',
            {'user_id': 1234567890}
            )
        request = mock.MagicMock()
        request.params = dict(signed_request=signed_request)
        request.context = SignedRequestContext(request)

        self.assertEqual(
            long(1234567890),
            fap.unauthenticated_userid(request)
            )

    def test_unauthenticated_bad_signature(self):
        from pyramid_facebook.lib import encrypt_signed_request
        from pyramid_facebook.security import (
            FacebookAuthenticationPolicy,
            SignedRequestContext,
            )

        # user id in signed request
        fap = FacebookAuthenticationPolicy('secret key')
        signed_request = encrypt_signed_request(
            'another secret key',
            {'user_id': 1234567890}
            )
        request = mock.MagicMock()
        request.params = dict(signed_request=signed_request)
        request.context = SignedRequestContext(request)

        self.assertIsNone(fap.unauthenticated_userid(request))

    def test_unauthenticated_bad_user_id(self):
        from pyramid_facebook.lib import encrypt_signed_request
        from pyramid_facebook.security import (
            FacebookAuthenticationPolicy,
            SignedRequestContext,
            )

        # user id in signed request
        fap = FacebookAuthenticationPolicy('secret key')
        signed_request = encrypt_signed_request(
            'secret key',
            {'user_id': "this is not a long"}
            )
        request = mock.MagicMock()
        request.params = dict(signed_request=signed_request)
        request.context = SignedRequestContext(request)

        self.assertIsNone(fap.unauthenticated_userid(request))


    def test_remember(self):
        from pyramid_facebook.security import FacebookAuthenticationPolicy

        fap = FacebookAuthenticationPolicy('secret key')

        request = mock.MagicMock()
        principal = "principal"
        kwargs = {"a": "a_val", "b": "b_val"}

        # FacebookAuthenticationPolicy does not support a remember/forget
        # mecanism, user is authenticated on each request
        self.assertListEqual([], fap.remember(request, principal, **kwargs))


    def test_forget(self):
        from pyramid_facebook.security import FacebookAuthenticationPolicy

        fap = FacebookAuthenticationPolicy('secret key')

        request = mock.MagicMock()

        # FacebookAuthenticationPolicy does not support a remember/forget
        # mecanism, user is authenticated on each request
        self.assertListEqual([], fap.forget(request))




# -*- coding: utf-8 -*-
from __future__ import absolute_import

import json
import logging
import urllib
from string import Template

from pyramid.events import subscriber
from pyramid.httpexceptions import HTTPFound, HTTPForbidden, HTTPOk
from pyramid.response import Response
from pyramid.view import (
    view_config,
    view_defaults
    )

from .security import SignedRequestContext, ViewCanvas

log = logging.getLogger(__name__)


js_redirect_tpl = """<html>
  <body>
    <script>
      window.top.location = "%(location)s";
    </script>
  </body>
</html>"""


class Base(object):
    "Base class for views and events"
    def __init__(self, context, request):
        # request and context as properties to prevent modification
        self._context = context
        self._request = request

    @property
    def context(self):
        """Route context which can be of 2 types:

        * :class:`~pyramid_facebook.security.SignedRequestContext`

        * :class:`~pyramid_facebook.security.FacebookCreditsContext`
        """
        return self._context

    @property
    def request(self):
        "Request object for this route."
        return self._request


class OauthAccept(Base):
    "Event sent when a user accepts app authentication."


class OauthDeny(Base):
    "Event sent when a user denies app authentication."


class DisputedOrder(Base):
    "Event sent when a user disputes an order."


class RefundedOrder(Base):
    "Event sent when a user got refunded for an order."


class PlacedItemOrder(Base):
    "Event sent when a user placed an item order."


class EarnedCurrencyOrder(Base):
    "Event sent when a user placed an currency order."


@view_defaults(route_name='facebook_canvas')
class FacebookCanvas(Base):
    """View for `facebook_canvas` route."""

    @view_config(context=HTTPForbidden)
    def prompt_authorize(self):
        """User is not logged in, view prompts user to authorize app:

        * User is not logged in
        * User is on apps.facebook.com/{namespace}
        """
        log.debug('prompt authorize')
        settings = self.request.registry.settings
        redirect_uri = urllib.quote_plus(
            "%s://apps.facebook.com%s" % (
                self.request.scheme,
                self.request.route_path('facebook_canvas_oauth')
                )
            )
        url = "%s/dialog/oauth/?client_id=%s&redirect_uri=%s&scope=%s" % (
            "https://www.facebook.com",
            settings['facebook.app_id'],
            redirect_uri,
            settings['facebook.scope'].replace(' ', '')
            )
        return Response(js_redirect_tpl % {'location': url})

    @view_config(permission=ViewCanvas)
    def canvas(self):
        """When users are logged in, they are authorized to view canvas.

        This view raises :py:exc:`exceptions.NotImplementedError`
        """
        raise NotImplementedError()


@view_defaults(route_name='facebook_canvas_oauth')
class FacebookOAuth(Base):

    def _redirect_to_canvas(self):
        p = {
            'location': "%s://apps.facebook.com%s" % (
                self.request.scheme,
                self.request.route_path('facebook_canvas')
                )
            }
        return Response(js_redirect_tpl % p)

    @view_config(permission=ViewCanvas)
    def oauth_accept(self):
        log.debug('user accepts')
        try:
            self.request.registry.notify(OauthAccept(self.context, self.request))
        except Exception as exc:
            log.error(
                'Oauth accept with context=%r, params=%r: %s',
                self.context,
                self.request.params,
                exc
                )
        return self._redirect_to_canvas()

    @view_config(context=HTTPForbidden, request_param="error")
    def oauth_deny(self):
        log.debug('user denies')
        try:
            self.request.registry.notify(OauthDeny(self.context, self.request))
        except Exception as exc:
            log.error(
                'Oauth deny with context=%r, params=%r: %s',
                self.context,
                self.request.params,
                exc
                )
        return self._redirect_to_canvas()


@view_config(route_name='facebook_payments_status_update_disputed')
def facebook_payments_status_update_disputed(context, request):
    try:
        log.debug('Order Status Update Disputed')
        request.registry.notify(DisputedOrder(context, request))
    except Exception as exc:
        log.error(
            'facebook_payments_status_update_disputed with context=%r, '
            'request.params=%r: %s',
            context,
            request,
            exc
            )
    return HTTPOk()


@view_config(route_name='facebook_payments_status_update_refunded')
def facebook_payments_status_update_refunded(context, request):
    try:
        log.debug('Order Status Update Refunded')
        request.registry.notify(RefundedOrder(context, request))
    except Exception as exc:
        log.error(
            'facebook_payments_status_update_refunded with context=%r, '
            'request.params=%r: %s',
            context,
            request,
            exc
            )
    return HTTPOk()


@view_config(
    route_name='facebook_payments_status_update_placed',
    permission=ViewCanvas,
    renderer='json'
    )
def facebook_payments_status_update_placed(context, request):
    if context.earned_currency_data:
        log.debug('Order Status Update - Earned App Currency')
        event_class = EarnedCurrencyOrder
    else:
        log.debug('Order Status Update - Placed Item Order')
        event_class = PlacedItemOrder
    try:
        request.registry.notify(event_class(context, request))
    except Exception as exc:
        log.error(
            'facebook_payments_status_update_placed with context=%r, '
            'request.params=%r: %s',
            context,
            request,
            exc
            )
        status = 'canceled'
    else:
        log.debug(
            'facebook_payments_status_update_placed settled %r',
            context.order_details
            )
        status = 'settled'
    return {
        'content':{
            'status': status,
            'order_id': context.order_details['order_id'],
            },
        'method':'payments_status_update',
        }



import unittest

import mock
from pyramid import testing
from pyramid.request import Request
from pyramid.response import Response
from pyramid.httpexceptions import HTTPOk


def _get_settings():
    return {
        'facebook.scope': 'mail, birthday',
        'facebook.app_id': '1234567890'
    }


class TestFacebookPaymentsCallback(unittest.TestCase):

    @mock.patch('pyramid_facebook.views.DisputedOrder')
    def test_facebook_payments_status_update_disputed(self, m_order):
        from pyramid_facebook.views import facebook_payments_status_update_disputed
        ctx = mock.Mock()
        req = mock.Mock()

        res = facebook_payments_status_update_disputed(ctx, req)

        req.registry.notify.assert_called_once_with(m_order.return_value)
        self.assertIsInstance(res, HTTPOk)

        # test exception
        req.reset_mock()
        req.registry.notify.side_effect = Exception('boom!')

        res = facebook_payments_status_update_disputed(ctx, req)

        req.registry.notify.assert_called_once_with(m_order.return_value)
        self.assertIsInstance(res, HTTPOk)

    @mock.patch('pyramid_facebook.views.RefundedOrder')
    def test_facebook_payments_status_update_refunded(self, m_order):
        from pyramid_facebook.views import facebook_payments_status_update_refunded
        ctx = mock.Mock()
        req = mock.Mock()

        res = facebook_payments_status_update_refunded(ctx, req)

        req.registry.notify.assert_called_once_with(m_order.return_value)
        self.assertIsInstance(res, HTTPOk)

        # test exception
        req.reset_mock()
        req.registry.notify.side_effect = Exception('boom!')

        res = facebook_payments_status_update_refunded(ctx, req)

        req.registry.notify.assert_called_once_with(m_order.return_value)
        self.assertIsInstance(res, HTTPOk)


    @mock.patch('pyramid_facebook.views.EarnedCurrencyOrder')
    def test_facebook_payments_status_update_placed_currency_app_order(self, m_order):
        from pyramid_facebook.views import facebook_payments_status_update_placed
        ctx = mock.MagicMock()
        req = mock.Mock()

        res = facebook_payments_status_update_placed(ctx, req)

        req.registry.notify.assert_called_once_with(m_order.return_value)

        expected = {
            'content':{
                'status': 'settled',
                'order_id': ctx.order_details['order_id']
                },
            'method': 'payments_status_update',
            }
        self.assertDictEqual(expected, res)

    @mock.patch('pyramid_facebook.views.PlacedItemOrder')
    def test_facebook_payments_status_update_placed_item_order(self, m_order):
        from pyramid_facebook.views import facebook_payments_status_update_placed
        ctx = mock.MagicMock()
        ctx.earned_currency_data = None
        req = mock.Mock()

        res = facebook_payments_status_update_placed(ctx, req)

        req.registry.notify.assert_called_once_with(m_order.return_value)

    @mock.patch('pyramid_facebook.views.EarnedCurrencyOrder')
    def test_facebook_payments_status_update_exception(self, m_order):
        from pyramid_facebook.views import facebook_payments_status_update_placed
        ctx = mock.MagicMock()
        req = mock.Mock()
        req.registry.notify.side_effect = Exception('boooom!')

        res = facebook_payments_status_update_placed(ctx, req)

        req.registry.notify.assert_called_once_with(m_order.return_value)

        expected = {
            'content':{
                'status': 'canceled',
                'order_id': ctx.order_details['order_id']
                },
            'method': 'payments_status_update',
            }
        self.assertDictEqual(expected, res)


class TestFacebookCanvas(unittest.TestCase):

    def test_init(self):
        from pyramid_facebook.views import FacebookCanvas
        request = mock.MagicMock()
        canvas = FacebookCanvas(request.context, request)
        self.assertEqual(request, canvas.request)
        self.assertEqual(request.context, canvas.context)

    def test_prompt_authorize(self):
        from pyramid_facebook.views import FacebookCanvas

        settings = _get_settings()

        request = mock.MagicMock()
        request.scheme = 'http'
        request.route_path.return_value = '/facebook/oauth'
        request.registry.settings = settings

        canvas = FacebookCanvas(request.context, request)

        result = canvas.prompt_authorize()

        request.route_path.assert_called_once_with('facebook_canvas_oauth')

        expected = """200 OK
Content-Type: text/html; charset=UTF-8
Content-Length: 234

<html>
  <body>
    <script>
      window.top.location = "https://www.facebook.com/dialog/oauth/?client_id=1234567890&redirect_uri=http%3A%2F%2Fapps.facebook.com%2Ffacebook%2Foauth&scope=mail,birthday";
    </script>
  </body>
</html>"""
        self.assertEqual(expected, str(result))

    def test_canvas(self):
        from pyramid_facebook.views import FacebookCanvas
        request = mock.MagicMock()
        canvas = FacebookCanvas(request.context, request)
        self.assertRaises(Exception, canvas.canvas)


class TestFacebookOauth(unittest.TestCase):

    def test_init(self):
        from pyramid_facebook.views import FacebookOAuth
        request = mock.MagicMock()
        canvas = FacebookOAuth(request.context, request)
        self.assertEqual(request, canvas.request)
        self.assertEqual(request.context, canvas.context)

    def test_oauth_accept(self):
        from pyramid_facebook.views import (
            FacebookOAuth,
            OauthAccept
            )
        settings = _get_settings()

        request = mock.MagicMock()
        request.scheme = 'http'
        request.route_path.return_value = '/facebook/'
        request.registry.settings = settings

        view = FacebookOAuth(request.context, request)

        result = view.oauth_accept()

        request.route_path.assert_called_once_with('facebook_canvas')
        self.assertTrue(request.registry.notify.called)

        self.assertIsInstance(
            request.registry.notify.call_args[0][0],
            OauthAccept
            )

        expected = """200 OK
Content-Type: text/html; charset=UTF-8
Content-Length: 126

<html>
  <body>
    <script>
      window.top.location = "http://apps.facebook.com/facebook/";
    </script>
  </body>
</html>"""
        self.assertEqual(expected, str(result))

    def test_oauth_deny(self):
        from pyramid_facebook.views import (
            FacebookOAuth,
            OauthDeny
            )
        settings = _get_settings()

        request = mock.MagicMock()
        request.scheme = 'http'
        request.route_path.return_value = '/facebook/'
        request.registry.settings = settings

        view = FacebookOAuth(request.context, request)

        result = view.oauth_deny()

        request.route_path.assert_called_once_with('facebook_canvas')
        self.assertTrue(request.registry.notify.called)

        self.assertIsInstance(
            request.registry.notify.call_args[0][0],
            OauthDeny
            )

        expected = """200 OK
Content-Type: text/html; charset=UTF-8
Content-Length: 126

<html>
  <body>
    <script>
      window.top.location = "http://apps.facebook.com/facebook/";
    </script>
  </body>
</html>"""
        self.assertEqual(expected, str(result))

    @mock.patch('pyramid_facebook.views.log')
    def test_oauth_exception(self, m_log):
        from pyramid_facebook.views import FacebookOAuth

        request = mock.MagicMock()
        request.registry.notify.side_effect= Exception('boooom!')

        view = FacebookOAuth(request.context, request)
        view.oauth_accept()
        self.assert_(m_log.error.called)

        m_log.error.reset_mock()
        view.oauth_deny()
        self.assert_(m_log.error.called)



# -*- coding: utf-8 -*-
import unittest

import mock


class TestLib(unittest.TestCase):

    def test_base64_encode_decode(self):
        from pyramid_facebook.lib import _base64_url_decode, _base64_url_encode

        msg = u'eyJhbGdvcml0aG0iOiJITUFDLVNIQTI1NiIsImlzc3VlZF9hdCI6MTI5NzExMDA0OCwidXNlciI6eyJsb2NhbGUiOiJlbl9VUyIsImNvdW50cnkiOiJjYSJ9fQ'
        data = u'{"algorithm":"HMAC-SHA256","issued_at":1297110048,"user":{"locale":"en_US","country":"ca"}}'

        self.assertEqual(data, _base64_url_decode(msg))
        self.assertEqual(msg, _base64_url_encode(data))
        self.assertEqual(
            msg,
            _base64_url_encode(
                _base64_url_decode(
                    _base64_url_encode(data)
                    )
                )
            )

    def test_encrypt_decrypt_request(self):
        from pyramid_facebook.lib import (
            encrypt_signed_request,
            decrypt_signed_request,
            )
        data = {
            u'issued_at': 1297110048,
            u'user': {
                u'locale': u'en_US',
                u'country': u'ca'
                },
            u'algorithm': u'HMAC-SHA256'
            }

        req = encrypt_signed_request('secret key', data)
        res = decrypt_signed_request('secret key', req)
        self.assertEqual(set(data.keys()), set(res.keys()))
        self.assertEqual(str(data), str(res))

    def test_decrypt_valid_signed_request_exception(self):
        from pyramid_facebook.lib import (
            FacebookSignatureException,
            decrypt_signed_request,
            encrypt_signed_request,
            )

        self.assertRaises(
            FacebookSignatureException,
            decrypt_signed_request,
            'secret_key',
            'without any dot'
            )

        self.assertRaises(
            FacebookSignatureException,
            decrypt_signed_request,
            'secret_key',
            'signature.'
            )

        self.assertRaises(
            FacebookSignatureException,
            decrypt_signed_request,
            'secret_key',
            '.payload'
            )

        self.assertRaises(
            FacebookSignatureException,
            decrypt_signed_request,
            'secret_key',
            'signature.payload'
            )

    @mock.patch('pyramid_facebook.lib.json')
    def test_decrypt_signed_request_exeption_json(self, m_json):
        from pyramid_facebook.lib import (
            _base64_url_encode,
            decrypt_signed_request,
            encrypt_signed_request,
            FacebookSignatureException
            )

        m_json.loads.return_value = []

        self.assertRaises(
            FacebookSignatureException,
            decrypt_signed_request,
            'secret key',
            "%s.payload" % _base64_url_encode("some input"),
            )


        m_json.loads.return_value = {}

        self.assertRaises(
            FacebookSignatureException,
            decrypt_signed_request,
            'secret key',
            "%s.payload" % _base64_url_encode("some input"),
            )


        m_json.loads.return_value = {'algorithm': 'NOT HMAC-SHA256'}

        self.assertRaises(
            FacebookSignatureException,
            decrypt_signed_request,
            'secret key',
            "%s.payload" % _base64_url_encode("some input"),
            )

    def test_decrypt_request_exception_wrong_signature(self):
        from pyramid_facebook.lib import (
            decrypt_signed_request,
            encrypt_signed_request,
            FacebookSignatureException
            )

        signed_request = encrypt_signed_request(
            'a secret key',
            {}
            )

        self.assertRaises(
            FacebookSignatureException,
            decrypt_signed_request,
            'another key',
            signed_request
            )

    def test_request_params_predicate(self):
        from pyramid_facebook.lib import request_params_predicate

        predicate = request_params_predicate('a', 'b', 'c', d=123, e='yo')

        request = mock.Mock()
        request.params = dict(a=1, b=2, c=3, d=123, e='yo')
        self.assert_(predicate(None, request))

        request.params = dict(a=1, b=2)
        self.assertFalse(predicate(None, request))

        request.params = dict(a=1, b=2, c=1, d=321, e='oy')
        self.assertFalse(predicate(None, request))



# -*- coding: utf-8 -*-
import json
import logging

from pyramid.authentication import CallbackAuthenticationPolicy
from pyramid.security import (
    IAuthenticationPolicy,
    Authenticated,
    Everyone,
    Allow
    )

from zope.interface import implementer

from .lib import (
    decrypt_signed_request,
    FacebookSignatureException
    )

__all__ = ['FacebookAuthenticationPolicy', 'SignedRequestContext', 'ViewCanvas',
    'FacebookCreditsContext']

log = logging.getLogger(__name__)

ViewCanvas = 'view_canvas'


class SignedRequestContext(object):
    "Security context for facebook signed request routes."

    def __init__(self, request):
        self.__acl__ = [
            (Allow, Authenticated, ViewCanvas),
        ]
        self._fb_data = None

    @property
    def facebook_data(self):
        """Contains facebook data provided in ``signed_request`` parameter
        decrypted with :func:`decrypt_signed_request <pyramid_facebook.lib.decrypt_signed_request>`.
        """
        return self._fb_data

    @facebook_data.setter
    def facebook_data(self, value):
        if self._fb_data is None:
            self._fb_data = value
        else:
            raise ValueError('Property can be set only once')

    def __repr__(self):
        return '<%s facebook_data=%r>' % (
            self.__class__.__name__,
            self._fb_data
            )


class FacebookCreditsContext(SignedRequestContext):
    "Context for facebook credits callback requests."

    @property
    def order_details(self):
        """Order details received in `facebook credits callback for payment
        status updates <http://developers.facebook.com/docs/credits/callback/#payments_status_update>`_."""
        if not hasattr(self, '_order_details'):
            self._order_details = json.loads(
                self.facebook_data['credits']['order_details']
                )
        return self._order_details

    @property
    def order_info(self):
        """Order info being the order information passed when the `FB.ui method
        <http://developers.facebook.com/docs/reference/javascript/FB.ui/>`_
        is invoked."""
        return self.facebook_data["credits"]["order_info"]

    @property
    def earned_currency_data(self):
        """Modified field received in `facebook credits callback for payment
        status update for earned app currency
        <http://developers.facebook.com/docs/credits/callback/#payments_status_update_earn_app_currency>`_."""
        if not hasattr(self, '_earned_currency_data'):
            data = self.order_details['items'][0]['data']
            if data:
                data = json.loads(data)['modified']
            self._earned_currency_data = data
        return self._earned_currency_data


@implementer(IAuthenticationPolicy)
class FacebookAuthenticationPolicy(CallbackAuthenticationPolicy):
    """A policy which authenticates user from ``signed_request`` parameter:

    * It decrypts ``signed_request`` as explained on `Facebook documentation
      <http://developers.facebook.com/docs/authentication/signed_request/>`_
      using :func:`decrypt_signed_request <pyramid_facebook.lib.decrypt_signed_request>`

    * It assigns :py:attr:`context.facebook_data <pyramid_facebook.security.SignedRequestContext.facebook_data>`
      with decrypted data.

    See `CallbackAuthenticationPolicy code <https://github.com/Pylons/pyramid/blob/master/pyramid/authentication.py#L35>`_
    for more info.
    """

    def __init__(self, secret_key, callback=None):
        log.debug('authentication policy init')
        self.secret_key = secret_key
        self.callback = callback

    def unauthenticated_userid(self, request):
        if request.params.get('signed_request') is None:
            return None

        context = request.context
        if context.facebook_data is None:
            # decrypt signed request once only
            try:
                context.facebook_data = decrypt_signed_request(
                    self.secret_key,
                    request.params["signed_request"]
                    )
            except FacebookSignatureException as e:
                context.facebook_data = dict()
                log.warn('Invalid signature %r', e)
                return None
        try:
            return long(context.facebook_data["user_id"])
        except KeyError:
            # user_id not in facebook_data => user has not authorized app
            return None
        except (TypeError, ValueError):
            log.warn('Invalid user id %r', context.facebook_data["user_id"])
        return None


    def remember(self, request, principal, **kwargs):
        return []


    def forget(self, request):
        return []



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import unittest
import webtest

from pyramid.config import Configurator


class TestController(unittest.TestCase):

    _app = None

    @property
    def app(self):
        if TestController._app is None:
            from pyramid_facebook import includeme
            from pyramid_facebook.tests import conf
            config = Configurator(settings=conf)
            config.include(includeme)
            TestController._app = webtest.TestApp(config.make_wsgi_app())
        return TestController._app




# -*- coding: utf-8 -*-
from __future__ import absolute_import

import json

from ...tests import get_signed_request, conf
from ..functionals import TestController

# Do not ask why facebook payment callback is that twisted:
# 1 - Read the docs https://developers.facebook.com/docs/credits/callback/
# 2 - if headache take aspirin and go back to point 1

USER_ID = 123
ORDER_ID = 987

credits_data = {
        'buyer': USER_ID,
        'receiver': USER_ID,
        'order_id': ORDER_ID,
        'order_info': '{"item_id":"1"}'
        }

request_data = {
    'user':{
        'country': 'us',
        'locale': 'en_US',
        'age': {
            'min': 21,
            'max': 45
            }
        },
    'user_id': USER_ID,
    'oauth_token':'AAAE0yTNwOxgBABXrDk9tNR0DoWyxwuLLNtpaJCRve8jbUZCr4YbxiIJqdwtHHZBt5uK8wV9ELHFfZBl9DmZAyvcXOuTZC9JZC22BCSfzky1gZDZD',
    'expires': 1325210400,
    'issued_at': 1325203762
    }

order_details = {
    'order_id': ORDER_ID,
    'buyer': USER_ID,
    'app': 9876,
    'receiver': USER_ID,
    'amount': 10,
    'time_placed': 1329243276,
    'update_time': 1329243277,
    'data':'',
    'items':[{
        'item_id': '0',
        'title':'100 Diamonds',
        'description': 'Spend Diamonds in dimonds game.',
        'image_url': 'http://image.com/diamonds.png',
        'product_url':'',
        'price': 10,
        'data':''
        }],
    'status': 'placed'}

earned_app_currency = {
    'modified':{
        'product': 'URL_TO_APP_CURR_WEBPAGE',
        'product_title': 'Diamonds',
        'product_amount': 3,
        'credits_amount':10
        }
    }


class TestFacebookCredits(TestController):
    """Test based on fb documentation:
    https://developers.facebook.com/docs/credits/callback/"""

    def test_payments_get_items(self):
        params = request_data.copy()
        params.update(credits=credits_data)
        params = {
            'signed_request': get_signed_request(**params),
            'method': 'payments_get_items'
            }
        params.update(credits_data)

        self.app.post(
            '/%s/credits' % conf['facebook.namespace'],
            params,
            status=404
            )

    def test_payments_status_update_placed_item_order(self):
        params = request_data.copy()
        params.update(
            {
            'credits':{
                'order_details': json.dumps(order_details.copy()),
                'status': 'placed',
                'order_id': ORDER_ID
                }
            })

        params = {
            'signed_request': get_signed_request(**params),
            'order_details': json.dumps(order_details.copy()),
            'order_id': ORDER_ID,
            'method': 'payments_status_update',
            'status': 'placed',
            }

        result = self.app.post(
            '/%s/credits' % conf['facebook.namespace'],
            params,
            )

        expected = {
            'content': {
                'status': 'settled',
                'order_id': 987
                },
            'method': 'payments_status_update'
            }
        self.assertDictEqual(expected, json.loads(result.body))


    def test_payments_status_update_earned_app_currency(self):
        params = request_data.copy()
        details = order_details.copy()
        details['items'][0]['data'] = json.dumps(earned_app_currency)
        params.update(
            {
            'credits':{
                'order_details': json.dumps(details),
                'status': 'placed',
                'order_id': ORDER_ID
                }
            })

        params = {
            'signed_request': get_signed_request(**params),
            'order_details': json.dumps(details),
            'order_id': ORDER_ID,
            'method': 'payments_status_update',
            'status': 'placed',
            }

        result = self.app.post(
            '/%s/credits' % conf['facebook.namespace'],
            params,
            )

        expected = {
            'content': {
                'status': 'settled',
                'order_id': 987
                },
            'method': 'payments_status_update'
            }
        self.assertDictEqual(expected, json.loads(result.body))



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import logging

import json
import venusian
from pyramid.view import view_config
from pyramid.config import Configurator
from pyramid.authorization import ACLAuthorizationPolicy

from .lib import request_params_predicate
from .security import FacebookAuthenticationPolicy, ViewCanvas

__all__ = [
    'includeme',
    'facebook_canvas'
    'facebook_payments_get_items',
    ]

log = logging.getLogger(__name__)


def includeme(config):
    """Executed when including ``pyramid_facebook``.

    ``pyramid_facebook`` setup:

    * ACL Authorization with :class:`~pyramid.authorization.ACLAuthorizationPolicy`
      using :meth:`config.set_authorization_policy <pyramid.config.Configurator.set_authorization_policy>`
    * Authentication with :class:`~pyramid_facebook.security.FacebookAuthenticationPolicy`
      using :meth:`config.set_authentication_policy <pyramid.config.Configurator.set_authentication_policy>`

    It adds routes which all requires :

    * a ``POST`` method, ``GET`` method will be considered as not found and
      will return a HTTP 404)
    * a ``signed_request`` parameter in body as defined on `Facebook
      documentation
      <http://developers.facebook.com/docs/authentication/signed_request/>`_.

    Routes added:

    * ``facebook_canvas`` associated to url ``/{namespace}/`` that musts be
      configured as canvas callback url in facebook application settings.

    * ``facebook_canvas_oauth`` associated to url ``/{namespace}/oauth`` for
      authentication.

    * ``facebook_payments_get_items`` associated to url ``/{namespace}/credits``.

    * ``facebook_payments_status_update_placed`` associated to url
      ``/{namespace}/credits``.

    * ``facebook_payments_status_update_disputed`` associated to url
      ``/{namespace}/credits``.

    * ``facebook_payments_status_update_refunded`` associated to url
      ``/{namespace}/credits``.
    """
    log.debug('config: %r', config)
    settings = config.registry.settings

    config.set_authentication_policy(
        FacebookAuthenticationPolicy(settings['facebook.secret_key'])
        )
    config.set_authorization_policy(ACLAuthorizationPolicy())

    facebook_path = '/%s' % settings['facebook.namespace']

    # Facebook canvas ##########################################################
    log.debug('Adding route facebook_canvas: %s/', facebook_path)
    config.add_route(
        'facebook_canvas',
        '%s/' % facebook_path,
        request_method='POST',
        request_param='signed_request',
        factory='pyramid_facebook.security.SignedRequestContext',
        )

    log.debug('Adding route facebook_canvas_oauth: %s/oauth', facebook_path)
    config.add_route(
        'facebook_canvas_oauth',
        '%s/oauth' % facebook_path,
        request_method='POST',
        request_param='signed_request',
        factory='pyramid_facebook.security.SignedRequestContext',
        )

    # Facebook credits #########################################################

    log.debug(
        'Adding route "facebook_payments_get_items": %s/credits',
        facebook_path
        )
    config.add_route(
        'facebook_payments_get_items',
        '%s/credits' % facebook_path,
        request_method='POST',
        custom_predicates=[
            request_params_predicate(
                'signed_request',
                'buyer',
                'receiver',
                'order_id',
                'order_info',
                method='payments_get_items',
                )
            ],
        factory='pyramid_facebook.security.FacebookCreditsContext',
        )

    update_predicate = request_params_predicate(
        'signed_request',
        'order_details',
        'order_id',
        method='payments_status_update',
        )

    log.debug(
        'Adding route "facebook_payments_status_update_placed": %s/credits',
        facebook_path
        )
    config.add_route(
        'facebook_payments_status_update_placed',
        '%s/credits' % facebook_path,
        request_method='POST',
        custom_predicates=[
            update_predicate,
            request_params_predicate(status='placed'),
            ],
        factory='pyramid_facebook.security.FacebookCreditsContext',
        )

    log.debug(
        'Adding route "facebook_payments_status_update_disputed": %s/credits',
        facebook_path
        )
    config.add_route(
        'facebook_payments_status_update_disputed',
        '%s/credits' % facebook_path,
        request_method='POST',
        custom_predicates=[
            update_predicate,
            request_params_predicate(status='disputed'),
            ],
        factory='pyramid_facebook.security.FacebookCreditsContext',
        )

    log.debug(
        'Adding route "facebook_payments_status_update_refunded": %s/credits',
        facebook_path
        )
    config.add_route(
        'facebook_payments_status_update_refunded',
        '%s/credits' % facebook_path,
        request_method='POST',
        custom_predicates=[
            update_predicate,
            request_params_predicate(status='refunded'),
            ],
        factory='pyramid_facebook.security.FacebookCreditsContext',
        )

    config.scan('pyramid_facebook.views')


class facebook_canvas(view_config):
    """Decorator that registers a view for the ``facebook_canvas`` route
    with the ``view_canvas`` permission.

    Accepts same arguments as :class:`~pyramid.view.view_config`::

        @facebook_canvas(renderer='canvas.mako')
        def canvas(context, request):
            return {
                'title': 'A great Facebook Game'
                }
    """
    def __init__(self, **kwargs):
        config = kwargs.copy()
        config.update({
            'permission': ViewCanvas,
            'route_name': 'facebook_canvas'
            })
        super(facebook_canvas, self).__init__(**config)


class facebook_payments_get_items(object):
    """Decorator to register the function to process facebook credits
    `payments_get_items <http://developers.facebook.com/docs/credits/callback/#payments_get_items>`_
    .

    Decorated function receives 2 positional parameters and 1 keyword argument:

    * ``context``: The :class:`~pyramid_facebook.security.SignedRequestContext`
      the request is associated with. ``context.facebook_data["user"]`` gives
      information about user's locale which would permit to return different
      languages.

    * ``request``: The request itself.

    It is possible to access `order_info` via :attr:`context.order_info
    <pyramid_facebook.security.FacebookCreditsContext.order_info>`:

    Decorated function must return a dictionary structured as::

        {
            # Required:
            "title":       "100 diamonds",
            "description": "100 shiny diamonds!",
            "price":       1000,
            "image_url":   "http://hadriendavid.com/images/100dimonds.png",

            # Optionnal (according to facebook doc):
            "item_id": "123",
            "data":    "whatever"
        }

    Example::

        @facebook_payments_get_items()
        def get_item(context, request):
            ...
            return {
                "title": a_title,
                "description": a_description,
                "price": a_price_in_facebook_credits,
                "image_url": an_image_url
                }
    """
    def __call__(self, wrapped):
        # as documented on http://docs.pylonsproject.org/projects/pyramid/en/1.3-branch/narr/hooks.html#registering-configuration-decorators
        venusian.attach(wrapped, self._register)
        return wrapped

    def _register(self, scanner, name, wrapped):
        """Register view to pyramid framework.
        """
        config = scanner.config
        config.add_view(
            view=self._decorate(wrapped),
            permission=ViewCanvas,
            route_name='facebook_payments_get_items',
            renderer='json',
            )

    def _validate(self, item_content):
        "Validate item content"
        if not isinstance(item_content, dict):
            raise TypeError('dict expected, received %s' % type(item_content))
        expected = {
            'title'      : (str, unicode,),
            'description': (str, unicode,),
            'price'      : (int,         ),
            'image_url'  : (str, unicode,),
            }
        for k, v in expected.iteritems():
            if type(item_content.get(k)) not in v:
                raise TypeError('Expected %r, received %r' % (
                    expected,
                    item_content
                    ))
        return item_content

    def _decorate(self, wrapped):
        def wrapper(context, request):
            info = context.facebook_data["credits"]["order_info"]
            result = self._validate(wrapped(context, request))
            return {
                "content": [result],
                "method": "payments_get_items"
                }
        return wrapper



# -*- coding: utf-8 -*-
import base64
import hashlib
import hmac
import json
import logging

log = logging.getLogger(__name__)

_signed_request_algorithm = 'HMAC-SHA256'
_dict_translate = dict(zip(map(ord, u'-_'), u'+/'))


class FacebookSignatureException(Exception):
    """Raised when ``signed_request`` is invalid.
    """
    pass


def _base64_url_encode(inp):
    """ Facebook base64 decoder based on `Sunil Arora's blog post
    <http://sunilarora.org/parsing-signedrequest-parameter-in-python-bas>`_.

    :param inp: input `str` to be encoded
    :return: `base64` encoded utf8 unicode data
    """
    return unicode(base64.b64encode(inp, '-_').strip('=').encode('utf8'))


def _base64_url_decode(inp):
    """ Facebook base64 decoder based on `Sunil Arora's blog post
    <http://sunilarora.org/parsing-signedrequest-parameter-in-python-bas>`_.

    :param inp: input `str` to be decoded
    :return: decoded `str`
    """
    padding_factor = (4 - len(inp) % 4) % 4
    inp += '=' * padding_factor
    return base64.b64decode(inp.translate(_dict_translate))


def encrypt_signed_request(secret_key, data):
    """Encrypts data the way facebook does for permit testing. Adds algorithm
    key to dict.

    :param secret_key: Facebook application' secret key.
    :param data: a dictionary of data to sign.
    :return: Signed request as defined by `Facebook documentation
             <http://developers.facebook.com/docs/authentication/signed_request/>`_

    """
    data = data.copy()
    data.update(algorithm=_signed_request_algorithm)

    payload = _base64_url_encode(json.dumps(data)).encode('utf8')
    signature = _base64_url_encode(
        hmac.new(
            secret_key.encode('utf8'),
            msg=payload.encode('utf8'),
            digestmod=hashlib.sha256
            ).digest()
        )
    return '%s.%s' % (signature, payload)


def decrypt_signed_request(secret_key, signed_request):
    """ Decrypts signed request sent by facebook. See `Facebook documentation
    <http://developers.facebook.com/docs/authentication/signed_request/>`_

    :param secret_key: Facebook application' secret key.
    :param signed_request: Signed request to be decrypted.
    :return: a dictionary with facebook parameters.
    """
    if signed_request.count('.') != 1:
        msg = 'signed_request badly formatted.'
        raise FacebookSignatureException(msg)

    signature, payload = signed_request.split('.')

    if not len(signature) > 0 or not len(payload) > 0:
        msg = 'Facebook signature and/or payload with null length.'
        raise FacebookSignatureException(msg)

    try:
        signature = _base64_url_decode(signature)
        data = json.loads(_base64_url_decode(payload))
    except:
        msg = 'Facebook signature or data base64 decoding failed.'
        raise FacebookSignatureException(msg)

    if not hasattr(data, 'get'):
        msg = 'Facebook signed request with invalid data.'
        raise FacebookSignatureException(msg)

    algorithm = data.get('algorithm')
    if not algorithm:
        msg = 'Facebook Signed Request without algorithm information.'
        raise FacebookSignatureException(msg)

    if algorithm.upper() != _signed_request_algorithm:
        msg = 'Facebook signature not encoded with HMAC-SHA256.'
        raise FacebookSignatureException(msg)

    expected_signature = hmac.new(
        secret_key.encode('utf8'),
        msg=payload.encode('utf8'),
        digestmod=hashlib.sha256
        ).digest()

    if signature != expected_signature:
        msg = 'Facebook signed request with invalid signature.'
        raise FacebookSignatureException(msg)

    return data


def request_params_predicate(*required_param_keys, **required_params):
    """Custom predicates to check if required parameter are in request
    parameters. Read :ref:`custom route predicates <pyramid:custom_route_predicates>`
    for more info::

        # /example?param1&param2=321
        config.add_route(
            'example',
            '/example',
            custom_predicates=[request_params_predicate('param1', param2=321)]
            )
    """
    required = set(required_param_keys)
    def predicate(info, request):
        if not required.issubset(set(request.params.keys())):
            return False
        for k, v in required_params.items():
            if v != request.params.get(k):
                return False
        return True
    return predicate
